import json
from typing import Any, Protocol

import yaml
import jinja2

try:
    import ujson

    UJSON_AVAILABLE = True
except ImportError:
    UJSON_AVAILABLE = False

try:
    import orjson

    ORJSON_AVAILABLE = True
except ImportError:
    ORJSON_AVAILABLE = False

jinja_env = jinja2.Environment(autoescape=True)


class ConfigDeserializer(Protocol):
    """
    Deserializers can be added to sovereign by creating a subclass
    and then specified in config:

    template_context:
      context:
          ...:
            protocol: ...
            serialization: <serializer name>
            path: ...
    """

    def deserialize(self, input: Any) -> Any:
        ...


class YamlDeserializer(ConfigDeserializer):
    def deserialize(self, input: Any) -> Any:
        return yaml.safe_load(input)


class JsonDeserializer(ConfigDeserializer):
    def deserialize(self, input: Any) -> Any:
        return json.loads(input)


class JinjaDeserializer(ConfigDeserializer):
    def deserialize(self, input: Any) -> Any:
        return jinja_env.from_string(input)


class StringDeserializer(ConfigDeserializer):
    def deserialize(self, input: Any) -> Any:
        return str(input)


class PassthroughDeserializer(ConfigDeserializer):
    def deserialize(self, input: Any) -> Any:
        return input


class UjsonDeserializer(ConfigDeserializer):
    def deserialize(self, input: Any) -> Any:
        if not UJSON_AVAILABLE:
            raise ImportError("Configured a UJSON deserializer but it's not installed")
        return ujson.loads(input)


class OrjsonDeserializer(ConfigDeserializer):
    def deserialize(self, input: Any) -> Any:
        if not ORJSON_AVAILABLE:
            raise ImportError(
                "Configured an ORJSON deserializer but it's not installed"
            )
        return orjson.loads(input)



import time
import uuid
import requests
from contextvars import ContextVar
from contextlib import nullcontext

from sovereign import config

_trace_id_ctx_var: ContextVar[str] = ContextVar("trace_id", default="")
_span_id_ctx_var: ContextVar[str] = ContextVar("span_id", default="")


def get_trace_id() -> str:
    return _trace_id_ctx_var.get()


def get_span_id() -> str:
    return _span_id_ctx_var.get()


def generate_128bit():
    return str(uuid.uuid4()).replace("-", "")


def generate_64bit():
    return generate_128bit()[:32]


def timestamp():
    return str(time.time()).replace(".", "")


TRACING = config.tracing
TRACING_DISABLED = not TRACING.enabled


class Tracer:
    def gen_id(self):
        if TRACING.trace_id_128bit:
            trace_id = generate_128bit()
        else:
            trace_id = generate_64bit()
        _trace_id_ctx_var.set(trace_id)
        return trace_id

    def __init__(self, span_name):
        if TRACING_DISABLED:
            return
        span_id = get_span_id()
        self.parent_span_id = None
        if span_id != "":
            # We are already inside a trace context
            self.parent_span_id = span_id
        self.trace_id = get_trace_id()
        self.span_id = self.gen_id()
        self.span_name = span_name

    def __enter__(self):
        if TRACING_DISABLED:
            return nullcontext()
        self.trace = {
            "traceId": self.trace_id,
            "id": self.span_id,
            "name": self.span_name,
            "timestamp": time.time(),
            "tags": TRACING.tags,
        }
        if self.parent_span_id:
            self.trace["parent_span_id"] = self.parent_span_id
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if TRACING_DISABLED:
            return
        self.trace["duration"] = time.time() - self.trace["timestamp"]
        self.submit()

    def submit(self):
        print(f"{self.span_name}: {self.trace['duration']}")
        try:
            url = f"{TRACING.collector}{TRACING.endpoint}"
            requests.post(url, json=self.trace)
        # pylint: disable=broad-except
        except Exception as e:
            print(f"Failed to submit trace: {self.trace}, Error:{e}")



import os
from contextvars import ContextVar
from importlib.metadata import version
from typing import Any, Mapping, Type

from fastapi.responses import JSONResponse
from pydantic import ValidationError
from starlette.templating import Jinja2Templates

from sovereign import dynamic_config
from sovereign.context import TemplateContext
from sovereign.logging.bootstrapper import LoggerBootstrapper
from sovereign.schemas import SovereignAsgiConfig, SovereignConfig, SovereignConfigv2
from sovereign.sources import SourcePoller
from sovereign.statistics import configure_statsd
from sovereign.utils.crypto.crypto import CipherContainer
from sovereign.utils.dictupdate import merge  # type: ignore
from sovereign.utils.resources import get_package_file

json_response_class: Type[JSONResponse] = JSONResponse
try:
    import orjson
    from fastapi.responses import ORJSONResponse

    json_response_class = ORJSONResponse
except ImportError:
    try:
        import ujson
        from fastapi.responses import UJSONResponse

        json_response_class = UJSONResponse
    except ImportError:
        pass


def parse_raw_configuration(path: str) -> Mapping[Any, Any]:
    ret: Mapping[Any, Any] = dict()
    for p in path.split(","):
        spec = dynamic_config.Loadable.from_legacy_fmt(p)
        ret = merge(obj_a=ret, obj_b=spec.load(), merge_lists=True)
    return ret


_request_id_ctx_var: ContextVar[str] = ContextVar("request_id", default="")


def get_request_id() -> str:
    return _request_id_ctx_var.get()


DIST_NAME = "sovereign"

__version__ = version(DIST_NAME)
config_path = os.getenv("SOVEREIGN_CONFIG", "file:///etc/sovereign.yaml")

html_templates = Jinja2Templates(
    directory=str(get_package_file(DIST_NAME, "templates"))
)

try:
    config = SovereignConfigv2(**parse_raw_configuration(config_path))
except ValidationError:
    old_config = SovereignConfig(**parse_raw_configuration(config_path))
    config = SovereignConfigv2.from_legacy_config(old_config)
asgi_config = SovereignAsgiConfig()
XDS_TEMPLATES = config.xds_templates()

logs = LoggerBootstrapper(config)
application_logger = logs.application_logger.logger

stats = configure_statsd(config=config.statsd)
poller = SourcePoller(
    sources=config.sources,
    matching_enabled=config.matching.enabled,
    node_match_key=config.matching.node_key,
    source_match_key=config.matching.source_key,
    source_refresh_rate=config.source_config.refresh_rate,
    logger=application_logger,
    stats=stats,
)

encryption_configs = config.authentication.encryption_configs
server_cipher_container = CipherContainer.from_encryption_configs(
    encryption_configs, logger=application_logger
)

template_context = TemplateContext(
    refresh_rate=config.template_context.refresh_rate,
    refresh_cron=config.template_context.refresh_cron,
    refresh_num_retries=config.template_context.refresh_num_retries,
    refresh_retry_interval_secs=config.template_context.refresh_retry_interval_secs,
    configured_context=config.template_context.context,
    poller=poller,
    encryption_suite=server_cipher_container,
    logger=application_logger,
    stats=stats,
)
poller.lazy_load_modifiers(config.modifiers)
poller.lazy_load_global_modifiers(config.global_modifiers)



import os
import importlib
from typing import Any, Protocol
from pathlib import Path
from importlib.machinery import SourceFileLoader

import requests

from sovereign.utils.resources import get_package_file_bytes

try:
    import boto3

    BOTO_IS_AVAILABLE = True
except ImportError:
    BOTO_IS_AVAILABLE = False


class CustomLoader(Protocol):
    """
    Custom loaders can be added to sovereign by creating a subclass
    and then in config:

    template_context:
      context:
          ...:
            protocol: <loader name>
            serialization: ...
            path: <path argument>
    """

    default_deser: str = "yaml"

    def load(self, path: str) -> Any:
        ...


class File(CustomLoader):
    default_deser = "passthrough"

    def load(self, path: str) -> Any:
        with open(path) as f:
            contents = f.read()
            try:
                return contents
            except FileNotFoundError:
                raise FileNotFoundError(f"Unable to load {path}")


class PackageData(CustomLoader):
    default_deser = "string"

    def load(self, path: str) -> Any:
        pkg, pkg_file = path.split(":")
        data = get_package_file_bytes(pkg, pkg_file)
        return data


class Web(CustomLoader):
    default_deser = "json"

    def load(self, path: str) -> Any:
        response = requests.get(path)
        response.raise_for_status()
        data = response.text
        return data


class EnvironmentVariable(CustomLoader):
    default_deser = "raw"

    def load(self, path: str) -> Any:
        data = os.getenv(path)
        if data is None:
            raise AttributeError(f"Unable to read environment variable {path}")
        return data


class PythonModule(CustomLoader):
    default_deser = "passthrough"

    def load(self, path: str) -> Any:
        if ":" in path:
            mod, fn = path.rsplit(":", maxsplit=1)
        else:
            mod, fn = path, ""
        imported = importlib.import_module(mod)
        if fn != "":
            return getattr(imported, fn)
        return imported


class S3Bucket(CustomLoader):
    default_deser = "raw"

    def load(self, path: str) -> Any:
        if not BOTO_IS_AVAILABLE:
            raise ImportError(
                "boto3 must be installed to load S3 paths. Use ``pip install sovereign[boto]``"
            )
        bucket, key = path.split("/", maxsplit=1)
        s3 = boto3.client("s3")
        response = s3.get_object(Bucket=bucket, Key=key)
        data = "".join([chunk.decode() for chunk in response["Body"]])
        return data


class PythonInlineCode(CustomLoader):
    default_deser = "passthrough"

    def load(self, path: str) -> Any:
        p = str(Path(path).absolute())
        loader = SourceFileLoader(p, path=p)
        return loader.load_module(p)


class Inline(CustomLoader):
    default_deser = "string"

    def load(self, path: str) -> Any:
        return path
